package com.dupliphyml;
/*
 * DupliphyML is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * DupliphyML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DupliphyML.  If not, see <http://www.gnu.org/licenses/>.
 */

import Alignments.Alignment;
import Alignments.Ambiguous;
import Alignments.DuplicationAlignment;
import Alignments.Site;
import Ancestors.AncestralJoint;
import Ancestors.AncestralJointDP;
import Ancestors.AncestralMarginal;
import Ancestors.AncestralMarginal.Result;
import Ancestors.AncestralMarginal.SiteResult;
import Exceptions.GeneralException;
import Exceptions.OutputException;
import Exceptions.UnexpectedError;
import Likelihood.StandardCalculator;
import Likelihood.StandardLikelihood;
import Likelihood.SiteLikelihood;
import Maths.Real;
import Maths.RealFactory;
import Maths.RealFactory.RealType;
import Maths.SquareMatrix;
import Optimizers.*;
import Parameters.Parameters;
import Models.Distributions;
import Models.DuplicationModelFactory;
import Models.RateCategory;
import Models.Model;
import Trees.Tree;
import Utils.PossibleSettings;
import Utils.SetSettings;
import Utils.SettingException;
import Utils.Settings;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Set;

/**
 * DupliphyML code.  This is pretty much a simple wrapper round GeLL
 * @author Daniel Money
 * @version 1.2
 */
public class DupliphyML
{
    /**
     * The main method.  See documentation for what command line inputs are
     * required.
     * @param args Command line arguements
     */
    public static void main(String[] args)
    {
	try
	{
	    Distributions.setMethod(Distributions.Calculation.REPEAT);
	    SquareMatrix.setExpMethod(SquareMatrix.Calculation.TAYLOR);   
            
            String aF, tF, model, output;
            String amF = "";
            String mF = "";
            RealType type = RealType.STANDARD_DOUBLE;
            int num;
            //Optimizer o = new GoldenSection(GoldenSection.ProgressLevel.PARAMETER);
            Optimizer o = new GoldenSection(GoldenSection.ProgressLevel.CALCULATION);
            boolean fixed = false;
            String ancestor = "Joint";
            boolean matrix = false;
                        
            if (args.length != 1)
            {
                if ((args.length < 5) || (args.length > 7))
                {
                    throw new SettingException("Wrong number of command line inputs");
                }
                aF = args[0];
                model = args[1];
                tF = args[2];
                output = args[3];
                num = Integer.parseInt(args[4]);
                if (args.length >= 6)
                {
                    mF = args[5];
                }
                if (args.length == 7)
                {
                    amF = args[6];
                }
            }
            else
            {
                PossibleSettings ps = getPossibleSettings();
                SetSettings ss = getSetSettings(args[0]);
                Settings settings = new Settings(ps,ss);
                
                aF = settings.getSetting("Families");
                tF = settings.getSetting("Tree");
                num = Integer.parseInt(settings.getSetting("Number"));
                model = settings.getSetting("Model");
                output = settings.getSetting("Output");
                mF = settings.getSetting("Missing");
                amF = settings.getSetting("Ambiguous");
                switch (settings.getSetting("Optimizer"))
                {
                    case "Nelder-Mead":
                        o = new NelderMead(NelderMead.DebugLevel.ON);
                        break;
                    case "Conjugate Gradient":
                        o = new ConjugateGradient(ConjugateGradient.ProgressLevel.PARAMETERS);
                        break;
                    case "Golden Section":
                        break;
                    default:
                        throw new SettingException("Unknown Optimizer");
                }
                switch (settings.getSetting("Real"))
                {
                    case "Standard":
                        break;
                    case "Small":
                        type = RealType.SMALL_DOUBLE;
                        break;
                    default:
                        throw new SettingException("Unknown Real Type");                    
                }
                fixed = settings.getSetting("FixedBranch").toLowerCase().startsWith("t");
                ancestor = settings.getSetting("Ancestral");
                matrix = settings.getSetting("Matrix").toLowerCase().startsWith("t");
            }

            SiteLikelihood.realType(type);
            AncestralJointDP.realType(type);

	    Alignment a = new DuplicationAlignment(new File(aF));
	    Tree t = Tree.fromFile(new File(tF));

            Model r;

	    int f= 0;
	    double n = num;
	    while (n > 12)
	    {
		n = n / 12;
		f++;
	    }
	    SquareMatrix.setForce(f);

            Ambiguous ambig = new Ambiguous(new HashMap<String,Set<String>>());
	    Alignment missing = null;
	    if (!amF.equals(""))
	    {
		ambig = Ambiguous.fromFile(new File(amF));
	    }
	    if (!mF.equals(""))
	    {
		missing = new DuplicationAlignment(new File(mF),ambig);
	    }

            Parameters p = new Parameters();
            switch (model)
            {
                case "Parsimony":
                    r = DuplicationModelFactory.Parsimony(p, num);
                    break;
                case "Parsimony+G":
                    r = DuplicationModelFactory.Parsimony_Gamma(p, num, 4);
                    break;
                case "BD_NoZero":
                    r = DuplicationModelFactory.BD_NoZero(p, num, fixed);
                    break;
                case "BD_NoZero+G":
                    r = DuplicationModelFactory.BD_NoZero_Gamma(p, num, 4, fixed);
                    break;
                case "BDIE":
                    r = DuplicationModelFactory.BDIE(p, num, fixed);
                    break;
                case "BDIE+G":
                    r = DuplicationModelFactory.BDIE_Gamma(p, num, 4, fixed);
                    break;
                case "BDI":
                    r = DuplicationModelFactory.BDI(p, num);
                    break;
                default:
                    r = DuplicationModelFactory.BDI_Gamma(p, num, 4);
                    break;
            }
            
            if (fixed)
            {
                p.addParameters(t.getParameters());
            }
            else
            {
                p.addParameters(t.getParametersForEstimation());
            }

	    StandardCalculator l = new StandardCalculator(r,a,t,missing);
            
	    StandardLikelihood like = o.maximise(l, p);
            System.out.println("Optimization complete");

            //Get the optimized parameters
            Parameters op = like.getParameters();
            
	    Tree lt = new Tree(t,op);

            Alignment anc;
            if (ancestor.equals("Marginal"))
            {
                AncestralMarginal am = new AncestralMarginal(r,a,lt);
                Result ancres = am.calculate(op);
                anc = ancres.getAlignment();
                
                File fap = new File(output + "_ancprob.txt");
                try
                {
                    PrintStream apout = new PrintStream(new FileOutputStream(fap));

                    List<String> internal = new ArrayList<>(t.getInternal());
                    Collections.sort(internal);

                    List<String> states = new ArrayList<>(r.getStates());
                    Collections.sort(states, new TextAsNumber());

                    apout.print("\t");
                    for (String st: states)
                    {
                        apout.print("\t" + st);
                    }
                    apout.println();
                    
                    for (Site s: a.getUniqueSites())
                    {
                        SiteResult sr = ancres.getSiteResult(s);
                        apout.println(s);
                        for (String i: internal)
                        {
                            apout.print("\t" + i);
                            for (String st: states)
                            {
                                apout.print(String.format("\t%6.4f",sr.getProbability(i, st).toDouble()));
                            }
                            apout.println();
                        }
                    }
                }
                catch (FileNotFoundException e)
                {
                    throw new OutputException(fap.getAbsolutePath(),"Unable to write out duplication alignment",e);
                }
            }
            else
            {
                AncestralJoint aj = AncestralJoint.newInstance(r,a,lt);
                anc = aj.calculate(op);
            }

	    op.toFile(new File(output + "_params.txt"), like.getLikelihood());
	    lt.toFile(new File(output + "_tree.ph"),true);
	    DuplicationAlignment.writeFile(anc, new File(output + "_desc.txt"));

            if (model.contains("+G"))
	    {
                File fr = new File(output + "_rate.txt");
                try
                {
                    PrintStream rout = new PrintStream(new FileOutputStream(fr));
                    for (int i = 0; i < a.getLength(); i++)
                    {    
                        Site s = a.getSite(i);
                        SiteLikelihood sl = like.getSiteLikelihood(s);
                        Real rate = RealFactory.getReal(type, 0.0);
                        for (RateCategory rc: r)
                        {
                            rate = rate.add(sl.getRateCategoryProbability(rc).multiply(rc.getTotalRate() * r.getScale()));
                        }
                        rout.println(i + "\t" + rate);
                    }
                    rout.close();
                }
                catch (FileNotFoundException e)
                {
                    throw new OutputException(fr.getAbsolutePath(),"Unable to write out rates",e);
                }
            }
            
            if (matrix)
            {
                File mof = new File(output + "_matrix.txt");
                try
                {
                    PrintStream mout = new PrintStream(new FileOutputStream(mof));
                    for (RateCategory rc: r.getRates())
                    {
                        mout.println(rc.getScaledMatrix());
                        mout.println();
                        mout.println(Arrays.toString(rc.getFreq()).replaceAll("(\\[|\\])", "").replace(", ", "\t"));
                    }
                }
                catch (FileNotFoundException e)
                {
                    throw new OutputException(mof.getAbsolutePath(),"Unable to write out matrix",e);
                }
            }
	}
	catch (GeneralException e)
	{
	    System.err.println(e.toString());
	}
	catch (SettingException e)
	{
	    System.err.println(e.toString());
	}

    }
    
    private static PossibleSettings getPossibleSettings()
    {
        try
        {
            PossibleSettings ps = new PossibleSettings();

            ps.addNeededSetting("Families");
            ps.addNeededSetting("Tree");
            ps.addNeededSetting("Model");            
            ps.addNeededSetting("Output");
            ps.addOptionalSetting("Number","75");
            ps.addOptionalSetting("Missing", "");
            ps.addOptionalSetting("Ambiguous", "");
            ps.addOptionalSetting("Optimizer", "Golden Section");
            ps.addOptionalSetting("Real", "Standard");
            ps.addOptionalSetting("FixedBranch", "false");
            ps.addOptionalSetting("Ancestral", "Joint");
            ps.addOptionalSetting("Matrix", "false");

            return ps;
        }
        catch (SettingException ex)
        {
            throw new UnexpectedError(ex);
        }
    }
     
    private static SetSettings getSetSettings(String sF) throws SettingException
    {
        try
        {
            return SetSettings.fromFile(new File(sF));
        }
        catch (FileNotFoundException e)
        {
            throw new SettingException("Settings file not found");
        }
        catch (IOException e)
        {
            throw new SettingException("Unable to read setting file");
        }
    }
    
    private static class TextAsNumber implements Comparator<String>
    {
        public int compare(String s1, String s2)
        {
            int i1 = Integer.parseInt(s1);
            int i2 = Integer.parseInt(s2);
            return Integer.compare(i1, i2);
        }
    }
}
